import json
import re
import six

from dateutil import parser

import kobold
from kobold import NotPresent
from kobold.hash_functions import (
        combine,
        acts_like_a_hash,
        acts_like_a_list)
from . import hints
from .hints import (
    JSONParsingHint,
    Base64Hint,
    ObjectAttrParsingHint,
    ObjectDictParsingHint,
    PickleParsingHint,
    ParsingHint)

pattern_type = getattr(re, '_pattern_type', None)
if pattern_type is None:
    pattern_type = getattr(re, 'Pattern')


def compare(expected, actual, type_compare=None):
    '''A wrapper around Compare.compare'''
    return Compare.compare(
            expected,
            actual,
            type_compare=type_compare)


class DontCare(object):
    '''Used as the "expected" argument in a comparison to mean "I don't 
       care what the 'actual' object is, as long as some rules hold."  
       By default, the rule is not_none_or_missing.
       
       Rules:

       not_none_or_missing: The second argument can be anything,
       as long as it is neither None nor kobold.NotPresent.
       
       list: The second argument can be anything, as long as it is
       an instance of list.
       
       json: The second argument may be any json-parseable string

       iso8601_datetime: The second argument must be able to be 
       evaluated as an ISO 8601 datetime string (eg. 
       2017-01-01T00:00:00)

       no_rules: The second argument may be anything - I literally
       do not care.
       '''

    
    def __init__(self,
                 rule='not_none_or_missing',
                 **kwargs):
        self.rule = rule
        self.options = kwargs
        self.validate()

    def validate(self):
        if self.rule == 'isinstance':
            if self.options.get('of_class') is None:
                raise kobold.ValidationError('isinstance dontcares must have an "of_class" option specified')
        

    def compare_with(self, other_thing):
        if self.rule == 'not_none_or_missing':
            return other_thing is not None
        elif self.rule == 'list':
            if not isinstance(other_thing, list):
                return False

            if self.options['length']:
                return len(other_thing) == self.options['length']
        elif self.rule == 'json':
            try:
                json.loads(other_thing)
                return True
            except:
                return False
        elif self.rule == 'iso8601_datetime':
            try:
                parser.parse(other_thing)
                return True
            except:
                return False
        elif self.rule == 'isinstance':
            return isinstance(other_thing, self.options['of_class'])
        elif self.rule is None or self.rule == 'no_rules':
            return True
        else:
            raise kobold.ValidationError('DontCare rule {} not recognized'.format(self.rule))
        

class ListDiff(list):
    def __init__(self, arr=None, display_type=list):
        if arr is None:
            arr = []
        self.with_positions = arr
        self.display_type = display_type
        val = [x for x in arr if x != '_']
        super(ListDiff, self).__init__(val)

    def display(self):
        to_display = []
        return self.display_type(self.with_positions)

    def append_match(self):
        self.with_positions.append('_')

    def append(self, value):
        super(ListDiff, self).append(value)
        self.with_positions.append(value)


class OrderedList(list):
    pass


class UnorderedList(list):
    pass

class StructuredString(object):
    '''An attempt to provide recursive comparison for strings'''

    def __init__(self, regex, arguments):
        self.regex = regex
        self.arguments = arguments



class Compare(object):
    @classmethod
    def compare(cls,
                expected,
                actual,
                type_compare=None):
        if type_compare is None:
            type_compare = {}
        elif isinstance(type_compare, str):
            type_compare = {'hash' : type_compare,
                            'list': 'full',
                            'ordered' : True}
        default_type_compare = {'hash' : 'full',
                                'list': 'full',
                                'ordered' : True}
        type_compare =\
            combine(default_type_compare, type_compare)
        if type_compare['ordered'] and type_compare['list'] == 'existing':
            raise kobold.ValidationError(
                'Ordered list compare must always be "full", not "existing"')

        if expected is DontCare:
            expected = DontCare()

        if isinstance(expected, DontCare):
            if expected.compare_with(actual):
                return 'match'
            else:
                return ("dontcare: %s" % expected.rule,
                        actual)
        elif (type(expected) == pattern_type and 
              isinstance(actual, six.string_types)):
            match = expected.match(actual)
            if match:
                return 'match'
            else:
                return ('regex: %s' % expected.pattern, actual)
        elif isinstance(expected, hints.ParsingHint):
            try:
                return cls.compare(
                        expected.payload,
                        expected.parse(actual),
                        type_compare)
            except kobold.InvalidMatch:
                return (expected, actual)
        elif (acts_like_a_hash(expected) and 
              acts_like_a_hash(actual)):
            return cls.hash_compare(expected, 
                                    actual, 
                                    type_compare)
        elif (isinstance(expected, tuple) and isinstance(actual, tuple)):
            return cls.list_compare(expected,
                                    actual,
                                    type_compare,
                                    iter_type=tuple)
        elif (acts_like_a_list(expected) and 
              acts_like_a_list(actual)):
            return cls.list_compare(expected, 
                                    actual,
                                    type_compare,
                                    iter_type=list)

        elif (isinstance(expected, StructuredString) and 
                isinstance(actual, six.string_types)):
            return cls.structured_string_compare(
                expected,
                actual,
                type_compare)

        else:
            if expected == actual:
                return 'match'
            else:
                return (expected, actual)

    @classmethod
    def hash_compare(cls,
                     expected,
                     actual,
                     type_compare={}):
        default_type_compare =\
            {'hash' : 'full',
             'dontcare_keys' : [],
             'ordered' : True}
        type_compare =\
            combine(default_type_compare, type_compare)
        if '__compare' in expected:
            compare_override = expected['__compare']
            if acts_like_a_hash(compare_override):
                type_compare = combine(default_type_compare, compare_override)
            else:
                type_compare['hash'] = compare_override
            expected = dict((k, v) for (k, v) in expected.items() if k != '__compare')

        expected_return = {}
        actual_return = {}

        if type_compare['hash'] == 'full':
            keys = set([])
            keys.update(expected.keys())
            keys.update(actual.keys())
        else:
            keys = set(expected.keys())

        for key in keys:
            if key in type_compare['dontcare_keys']:
                result = cls.compare(DontCare, 
                                     actual.get(key, kobold.NotPresent),
                                     type_compare)
            else:
                result = cls.compare(expected.get(key, kobold.NotPresent),
                                     actual.get(key, kobold.NotPresent),
                                     type_compare)

            if result != 'match':
                expected_sub, actual_sub = result
                expected_return[key] = expected_sub
                actual_return[key] = actual_sub

        if len(expected_return) == 0:
            return 'match'
        else:
            return (expected_return, actual_return)

    @classmethod
    def ordered_list_compare(cls,
                             expected,
                             actual,
                             type_compare,
                             iter_type=list):
        expected_elements = ListDiff(display_type=iter_type)
        actual_elements = ListDiff(display_type=iter_type)

        for i in range(max(len(expected), len(actual))):
            if len(expected) > i:
                expected_value = expected[i]
            else:
                expected_value = kobold.NotPresent
            if len(actual) > i:
                actual_value = actual[i]
            else:
                actual_value = kobold.NotPresent
            match = True
            result = cls.compare(expected_value,
                                 actual_value,
                                 type_compare)
            if result == 'match':
                expected_elements.append_match()
                actual_elements.append_match()
            else:
                expected_sub, actual_sub = result
                expected_elements.append(expected_sub)
                actual_elements.append(actual_sub)

        if (len(expected_elements) == 0 and
            len(actual_elements) == 0):
            return 'match'
        else:
            return (expected_elements.display(),
                    actual_elements.display())

    @classmethod
    def unordered_list_compare(cls,
                               expected,
                               actual,
                               type_compare,
                               iter_type=list):
        # Make a list of all the indexes of the "expected" list 
        # and the "actual" list.  
        # Iterate through the "expected" list.  For each item,
        # try to find a corresponding match in the "actual" list
        # (by iterating through that - n^2 style).
        # If a match is found, remove the corresponding indexes
        # from the "expected" and "actual" lists.  What we're left
        # with is two lists of missing indexes (one from the expected,
        # one from the actual).
        
        missing_expected_indexes = list(range(len(expected)))
        missing_actual_indexes = list(range(len(actual)))
        expected_index_index = 0
        while expected_index_index < len(missing_expected_indexes):
            expected_index = missing_expected_indexes[expected_index_index]
            expected_element = expected[expected_index]
            actual_index_index = 0
            while actual_index_index < len(missing_actual_indexes):
                actual_index = missing_actual_indexes[actual_index_index]
                actual_element = actual[actual_index]
                result = cls.compare(expected_element, 
                                     actual_element, 
                                     type_compare)
                if result == 'match':
                    missing_expected_indexes.pop(expected_index_index)
                    missing_actual_indexes.pop(actual_index_index)
                    expected_index_index -= 1
                    actual_index_index -= 1
                    break
                actual_index_index += 1
            expected_index_index += 1

       
        # The remaining elements in the expected and actual
        # lists (the elements that didn't have a partner in the
        # other list) are all still "full".  My theory (unsubstantiated)
        # is that it is more helpful for them to be displayed "diffed".
        # The question is, diffed with what.  Since this is an unordered
        # comparison, there's no obviously right answer to that question.
        # I figure, let's diff the missing elements from the "expected" 
        # list with the missing elements from the "actual" list in order.
        # That should at least give us friendlier output.
        # So, this section gets us the ordered diffs of the remaining
        # elements
        max_len = max(len(expected), len(actual))
        displayed_expecteds = [None] * max_len
        displayed_actuals = [None] * max_len

        for i in range(max(len(missing_expected_indexes),
                           len(missing_actual_indexes))):
            if i < len(missing_expected_indexes):
                missing_expected_index = missing_expected_indexes[i]
                missing_expected = expected[missing_expected_index]
            else:
                missing_expected_index = None
                missing_expected = kobold.NotPresent

            if i < len(missing_actual_indexes):
                missing_actual_index = missing_actual_indexes[i]
                missing_actual = actual[missing_actual_index]
            else:
                missing_actual_index = None
                missing_actual = kobold.NotPresent

            displayed_expected = cls.display(
                    missing_expected,
                    missing_actual)
            displayed_actual = cls.display(
                    missing_actual,
                    missing_expected)


            if missing_expected_index is not None:
                displayed_expecteds[missing_expected_index] = displayed_expected

            if missing_actual_index is not None:
                displayed_actuals[missing_actual_index] = displayed_actual

        # For each element of the expected and actual lists,
        # if there was a match, just append the "match" character
        # (_).  If there wasn't a match, append the in-order diff
        # from above.
        expected_return = ListDiff(display_type=iter_type)
        actual_return = ListDiff(display_type=iter_type)
        for i in range(len(expected)):
            if i in missing_expected_indexes:
                expected_return.append(displayed_expecteds[i])
            else:
                expected_return.append_match()

        for j in range(len(actual)):
            if j in missing_actual_indexes:
                actual_return.append(displayed_actuals[j])
            else:
                actual_return.append_match()

        if type_compare['list'] == 'full':
            if (len(expected_return) == 0 and
                len(actual_return) == 0):
                return 'match'
            else:
                return (expected_return.display(),
                        actual_return.display())
        elif type_compare['list'] == 'existing':
            if len(expected_return) == 0:
                return 'match'
            else:
                return (expected_return.display(),
                        actual_return.display())
        else:
            raise NotImplementedError(
                'Invalid value for list match type_compare '
                'setting: {}'.format(
                    type_compare['list']))
                                 
    @classmethod
    def list_compare(cls,
                     expected,
                     actual,
                     type_compare,
                     iter_type=list):
        default_type_compare =\
            {'hash' : 'full',
             'ordered' : True}

        type_compare =\
            combine(default_type_compare, type_compare)

        if type(expected) == set and type(actual) == set:
            isset = True
            expected = list(expected)
            actual = list(actual)
        else:
            isset = False

        if (isinstance(expected, OrderedList) or
                (type_compare['ordered'] and 
                    not isset and 
                    not isinstance(expected, UnorderedList)
                )
            ):
            ret = cls.ordered_list_compare(expected,
                                           actual,
                                           type_compare,
                                           iter_type=iter_type)
        else:
            ret = cls.unordered_list_compare(expected,
                                             actual,
                                             type_compare,
                                             iter_type=iter_type)
        if ret == 'match' or not isset:
            return ret
        else:
            ret_exp, ret_act = ret
            ret_exp = [x for x in ret_exp if x != '_']
            ret_act = [x for x in ret_act if x != '_']
            return (set(ret_exp), set(ret_act))


    # These "display" functions are used by the unordered list comparison
    # for intelligently displaying unordered diffs of lists
    @classmethod
    def display(cls, element, other_element):
        if element is DontCare:
            element = DontCare()
        if isinstance(element, DontCare):
            return "dontcare: %s" % element.rule
        elif acts_like_a_hash(element) and acts_like_a_hash(other_element):
            return cls.display_hash(element, other_element)
        elif isinstance(element, tuple) and isinstance(other_element, tuple):
            return cls.display_list(element, other_element, iter_type=tuple)
        elif acts_like_a_list(element) and acts_like_a_list(other_element):
            return cls.display_list(element, other_element, iter_type=list)
        elif type(element) == pattern_type:
            return 'regex: %s' % element.pattern
        elif isinstance(other_element, hints.ParsingHint):
            try:
                return cls.display(
                    other_element.parse(element),
                    other_element.payload)
            except kobold.InvalidMatch:
                return cls.display(
                        element,
                        other_element.payload)
        elif isinstance(element, hints.ParsingHint):
            try:
                return cls.display(
                        element.payload,
                        element.parse(other_element))
            except kobold.InvalidMatch:
                return cls.display(
                        element.payload,
                        other_element)
        else:
            return element

    @classmethod
    def display_hash(cls, one_hash, other_hash):
        display_hash = {}
        for key in one_hash.keys():
            display_hash[key] = cls.display(one_hash.get(key), other_hash.get(key))

        return display_hash

    @classmethod
    def display_list(cls, one_list, other_list, iter_type=list):
        max_len = max(len(one_list), len(other_list))
        display_list = []

        for i in range(max_len):
            if i > len(one_list) - 1:
                element = kobold.NotPresent
            else:
                element = one_list[i]

            if i > len(other_list) - 1:
                other_element = kobold.NotPresent
            else:
                other_element = other_list[i]

            display_list.append(cls.display(element, other_element))
        return iter_type(display_list)

    @classmethod
    def structured_string_compare(cls, expected, actual, type_compare={}):
        default_type_compare =\
            {'hash' : 'full',
             'dontcare_keys' : [],
             'ordered' : True}
        type_compare =\
            combine(default_type_compare, type_compare)

        match = expected.regex.match(actual)
        if match:
            groups = match.groups()
            return cls.compare(
                expected.arguments,
                match.groups(),
                type_compare=type_compare)
        else:
            return (
                'structured string regex: {}'.format(
                    expected.regex.pattern),
                actual)
                        

        






import re
import unittest

from kobold import (
        compare,
        doubles,
        swap)

class TestException(Exception):
    pass

def test_function(*args, **kwargs):
    return 'test_function'

class Host(object):
    def __init__(self, *args, **kwargs):
        for (name, value) in kwargs.items():
            setattr(self, name, value)

    def test(self, *args, **kwargs):
        return 'original'

class TestRoutableStubFunction(unittest.TestCase):
    def setUp(self):
        self.safe_swap = swap.SafeSwap()

    def tearDown(self):
        self.safe_swap.rollback()

    def test_kwarg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'kwarg' : 1},
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition={'kwarg' : 3},
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(arg=1, kwarg=1))
        self.assertEqual(4, stub_function(arg=1, kwarg=3))

    def test_arg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=(1,),
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition=(3,),
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(1))
        self.assertEqual(4, stub_function(3))

    def test_self_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={
                'self': {'a': 1}},
            stub_type='value',
            stub_value=1)
        stub_function.add_route(
            condition={
                'self': {'a': 2}},
            stub_type='value',
            stub_value=2)
        self.safe_swap.swap(
            Host,
            'test',
            stub_function,
            default_original=True)

        host1 = Host(a=1)
        host2 = Host(a=2)
        self.assertEqual(
                1,
                host1.test())
        self.assertEqual(
                2,
                host2.test())

    def test_arg_and_kwarg_routing(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'args' : (1,),
                       'kwargs' : {'kwarg' : 1}},
            stub_type='value',
            stub_value=2)
        stub_function.add_route(
            condition={'args' : (3,),
                       'kwargs' : {'kwarg' : 3}},
            stub_type='value',
            stub_value=4)

        self.assertEqual(2, stub_function(1, kwarg=1))
        self.assertEqual(4, stub_function(3, kwarg=3))



    def test_kwarg_with_pattern(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition={'kwarg' : re.compile('.*pattern1.*')},
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function('_', kwarg='something_with_pattern1'))
        self.assertRaises(
                doubles.StubRoutingException,
                stub_function,
                arg='_',
                kwarg='something_without_the_pattern')

    def test_arg_with_pattern(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=(re.compile('.*pattern1.*'),),
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function('something_with_pattern1', kwarg='_'))
        self.assertRaises(
            doubles.StubRoutingException,
            stub_function,
            'something_without_the_pattern',
            kwarg='_')

    def test_arg_is_dict(self):
        stub_function = doubles.RoutableStubFunction()
        stub_function.add_route(
            condition=({'a': 1},),
            stub_type='value',
            stub_value=1)

        self.assertEqual(1, stub_function({'a': 1}))

    def test_rotate(self):
        stub_function = doubles.RoutableStubFunction()
        route = doubles.RotatingRoute(
                    condition=('a',),
                    routes=[
                     ('value', 1),
                     ('value', 2)])
                     
        stub_function.add_route(route=route)

        self.assertEqual(1, stub_function('a'))
        self.assertEqual(2, stub_function('a'))


class TestStubFunction(unittest.TestCase):
    def test_raises(self):
        e = TestException()
        stub_function = doubles.StubFunction(raises=e)
        self.assertRaises(
            TestException,
            stub_function)

    def test_calls(self):
        stub_function = doubles.StubFunction(calls=test_function)
        self.assertEqual('test_function', stub_function())

    def test_returns(self):
        stub_function = doubles.StubFunction(returns=1)
        self.assertEqual(1, stub_function())

class TestSpyFunction(unittest.TestCase):
    def test_reset(self):
        spy_function = doubles.SpyFunction(returns=1)
        self.assertEqual(1, spy_function(1))
        self.assertEqual(1, spy_function("apple"))
        self.assertEqual(1, spy_function(keyword="orange"))
        spy_function.reset()

        self.assertEqual([],
                          spy_function.calls)

    def test_spy(self):
        spy_function = doubles.SpyFunction(returns=1)
        self.assertEqual(1, spy_function(1))
        self.assertEqual(1, spy_function("apple"))
        self.assertEqual(1, spy_function(keyword="orange"))

        self.assertEqual([((1,), {}), (("apple",), {}), ((), {'keyword' : 'orange'})],
                          spy_function.calls)


class TestRoutableSpyFunction(unittest.TestCase):
    def test_delegation(self):
        spy_function = doubles.SpyFunction(
                stub_function_factory=doubles.RoutableStubFunction)
        spy_function.add_route(
                condition=(1, 1),
                stub_type='value',
                stub_value=1)
        self.assertEqual(1, spy_function(1, 1))
        self.assertRaises(doubles.StubRoutingException,
                          spy_function,
                          1, 2)
        self.assertEqual([((1, 1), {}), ((1, 2), {})], spy_function.calls)



import kobold
import six


def project(hash_in, attributes):
    '''
    Given a dict and a list of keys, return a
    dict that only includes the keys in the list
    '''

    return dict(
        (key, value) for (key, value) 
        in hash_in.items() if key in attributes)

def merge(default, to_mutate):
    '''
    Mutate a dict (to_mutate) with defaults from the first.
    Obviously, if to_mutate already has a key, don't override it.
    Return the mutated dict.
    '''

    for key, value in default.items():
        to_mutate.setdefault(key, value)

    return to_mutate 


def combine(default, extra):
    '''
    Take two dicts and create a third dict with all the keys
    from both.  If both dicts share a key, the second dict
    overrides the first.  Return the new dict.
    '''

    new = {}
    for key, value in default.items():
        new[key] = value

    for key, value in extra.items():
        new[key] = value

    return new

def deep_combine(default, extra):
    keys = set()
    keys.update(default.keys())
    keys.update(extra.keys())
    new = {}
    for key in keys:
        default_for_key = default.get(key, kobold.NotPresent)
        extra_for_key = extra.get(key, kobold.NotPresent)
        if (isinstance(default_for_key, dict) and 
            isinstance(extra_for_key, dict)):
            new[key] = deep_combine(
                    default_for_key,
                    extra_for_key)
        else:
            new[key] = default_for_key
            if extra_for_key is not kobold.NotPresent:
                new[key] = extra_for_key
    return new


def acts_like_a_hash(candidate):
    return hasattr(candidate, 'items')

def acts_like_a_list(candidate):
    return (hasattr(candidate, '__iter__') and 
            hasattr(candidate, '__len__') and
            not isinstance(candidate, six.string_types) and
            not isinstance(candidate, dict))


# Based on https://stackoverflow.com/a/1151686/2619588'''
class HashableDict(dict):
    def __key(self):
        return tuple((k,self[k]) for k in sorted(self))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if not isinstance(other, dict):
            return False

        if not isinstance(other, HashableDict):
            other = self.from_dict(other)
        return self.__key() == other.__key()

    @classmethod
    def from_dict(self, dictionary):
        return HashableDict(**dictionary)

class HashableList(list):
    def __key(self):
        return tuple(self)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if not isinstance(other, list):
            return False

        if not isinstance(other, HashableList):
            other = self.from_list(other)
        return self.__key() == other.__key()

    @classmethod
    def from_list(self, original_list):
        new_list = HashableList()
        new_list.extend(original_list)
        return new_list

def make_hashable(data_structure):
    if acts_like_a_list(data_structure):
        new_data_structure = HashableList()
        for element in data_structure:
            new_data_structure.append(make_hashable(element))
        return new_data_structure
    elif acts_like_a_hash(data_structure):
        new_data_structure = HashableDict()
        for key, value in data_structure.items():
            new_key = make_hashable(key)
            new_value = make_hashable(value)
            new_data_structure[new_key] = new_value
        return new_data_structure
    else:
        return data_structure






class NotPresent(object):
    '''Used to explicitly specify that something isn't present,
       or wasn't supplied'''
    pass


class InvalidMatch(Exception):
    pass


class ValidationError(Exception):
    pass


from . import (
    assertions,
    compare,
    doubles,
    hash_functions,
    response)



import unittest
from kobold import assertions, compare


class ObjectThing(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

class TestCompare(unittest.TestCase):
    def test_empty_hashes(self):
        self.assertEqual('match', compare.compare({}, {}))

    def test_equal_hashes(self):
        self.assertEqual('match', compare.compare({'a' : 1}, {'a' : 1}))

    def test_distinct_keys(self):
        self.assertEqual(
                ({'a' : 1, 'b' : compare.NotPresent}, {'a' : compare.NotPresent, 'b' : 2}), 
                compare.compare(
                    {'a' : 1}, 
                    {'b' : 2}))

    def test_distinct_keys_only_existing(self):
        self.assertEqual(
                ({'a' : 1}, {'a' : compare.NotPresent}), 
                compare.compare(
                    {'a' : 1}, 
                    {'b' : 2},
                    type_compare='existing'))

    def test_intersecting_keys(self):
        self.assertEqual(
                ({'a' : 1, 'c' : compare.NotPresent}, {'a' : compare.NotPresent, 'c' : 3}),
                compare.compare(
                    {'a' : 1, 'b' : 2},
                    {'b' : 2, 'c' : 3}))

    def test_same_keys_different_values(self):
        self.assertEqual(
                ({'a' : 1, 'b' : 2}, {'a' : 2, 'b' : 1}),
                compare.compare(
                    {'a' : 1, 'b' : 2},
                    {'a' : 2, 'b' : 1}))

    def test_ordered_lists(self):
        self.assertEqual(
                ([1, 2], [2, 1]),
                compare.compare(
                    [1, 2],
                    [2, 1]))

    def test_unordered_lists(self):
        self.assertEqual(
                'match',
                compare.compare(
                    [1, 2],
                    [2, 1],
                    type_compare={'ordered' : False}))

    def test_one_element_off(self):
        self.assertEqual(
                (['_', 2, '_'], ['_', 4, '_']),
                compare.compare(
                    [1, 2, 3],
                    [1, 4, 3]))

    def test_parseable_json_dict(self):
        expected = compare.hints.JSONParsingHint({'a' : 1})
        actual = '{"a" : 1}'
        self.assertEqual(
                'match',
                compare.compare(expected, actual))

    def test_parseable_json_list(self):
        expected = compare.hints.JSONParsingHint([1, 2, 3])
        actual = '[1, 2, 3]'
        self.assertEqual(
                'match',
                compare.compare(expected, actual))

    def test_parseable_json_mismatch(self):
        expected = compare.hints.JSONParsingHint({'a' : 1})
        actual = '{"a" : "1"}'
        self.assertEqual(
                ({'a' : 1}, {'a' : '1'}),
                compare.compare(expected, actual))

    def test_expected_list_longer_than_actual(self):
        expected = [{'a' : 1}, {'b' : 2}]
        actual = [{'a' : 1}]
        self.assertEqual(
                (['_', {'b' : 2}], ['_', compare.NotPresent]),
                compare.compare(expected, actual))

    def test_expected_list_shorter_than_actual(self):
        expected = [{'a' : 1}]
        actual = [{'a' : 1}, {'b' : 2}]
        self.assertEqual(
                (['_', compare.NotPresent], ['_', {'b' : 2}]),
                compare.compare(expected, actual))

    def test_compare_dict_with_object_dict(self):
        expected = {'a' : 1, 'b' : 2, 'c' : 3}
        actual = ObjectThing(a=1, b=2, c=3)
        self.assertEqual(
                'match',
                compare.compare(
                    compare.hints.ObjectDictParsingHint(expected),
                    actual))

    def test_compare_notpresent_with_object_dict(self):
        expected = {'a' : 1, 'b' : 2, 'c' : 3}
        assertions.assert_match(
                ({'a' : 1, 'b' : 2, 'c' : 3},
                 compare.NotPresent),
                compare.compare(
                    compare.hints.ObjectDictParsingHint(expected),
                    compare.NotPresent))

    def test_compare_dict_with_object_dict_mismatch(self):
        expected = {'a' : 1, 'b' : 2, 'c' : 4}
        actual = ObjectThing(a=1, b=2, c=3)
        self.assertEqual(
                ({'c' : 4}, {'c' : 3}),
                compare.compare(
                    compare.hints.ObjectDictParsingHint(expected),
                    actual))

    def test_list_of_objects(self):
        expected = {'a' : 1, 'b' : 2, 'c' : 3}
        actual = [ObjectThing(a=1, b=2, c=4)]
        self.assertEqual(
                ([{'c' : 3}], [{'c' : 4}]),
                compare.compare(
                    [compare.hints.ObjectDictParsingHint(expected)],
                    actual))
    
    def test_list_of_objects_unordered(self):
        expected = [
                compare.hints.ObjectDictParsingHint({'a' : 1}),
                compare.hints.ObjectDictParsingHint({'a' : 2}),
                compare.hints.ObjectDictParsingHint({'a' : 3})]
        actual = [
                ObjectThing(a=4),
                ObjectThing(a=5),
                ObjectThing(a=6)]

        self.assertEqual(
                ([{'a' : 1}, {'a' : 2}, {'a' : 3}],
                 [{'a' : 4}, {'a' : 5}, {'a' : 6}]),
                compare.compare(
                    expected, 
                    actual,
                    type_compare={'hash' : 'existing',
                                  'ordered' : False}))

    def test_unordered_comparison_object_dict_one_item(self):
        expected = [
                compare.hints.ObjectDictParsingHint(
                    dict(a=1, b=2, c=3, d=4, e=5))]
        actual = [
                ObjectThing(a=1, b=3, c=3, d=4, e=5)]
        self.assertEqual(
                ([dict(a=1, b=2, c=3, d=4, e=5)],
                 [dict(a=1, b=3, c=3, d=4, e=5)]),
                compare.compare(
                    expected,
                    actual,
                    type_compare={'hash' : 'existing',
                                  'ordered' : False}))

    def test_unordered_comparison_object_dict_multiple_items(self):
        expected = [
                compare.hints.ObjectDictParsingHint(dict(a=1, b=2)),
                compare.hints.ObjectDictParsingHint(dict(a=2, b=3)),
                compare.hints.ObjectDictParsingHint(dict(a=3, b=4))]
        actual = [ObjectThing(a=2, b=3),
                  ObjectThing(a=2, b=1),
                  ObjectThing(a=3, b=2)]
        self.assertEqual(
                ([dict(a=1, b=2), '_', dict(a=3, b=4)],
                 ['_', dict(a=2, b=1), dict(a=3, b=2)]),
                compare.compare(
                    expected,
                    actual,
                    type_compare={'hash' : 'existing',
                                  'ordered' : False}))

    def test_two_sets(self):
        expected = set([1, 2, 3])
        actual = set([1, 3, 4])
        self.assertEqual(
            (set([2]), set([4])),
            compare.compare(expected, actual))


    def test_tuples(self):
        expected = (1, 2, 3, 4)
        actual = (1, 2, 4, 4)
        self.assertEqual(
            (('_', '_', 3, '_'), ('_', '_', 4, '_')),
            compare.compare(expected, actual))

    def test_list_of_tuples(self):
        expected = [(1, 2), (3, 4), (5, 6)]
        actual = [(1, 2), (3, 5), (5, 6)]
        self.assertEqual(
            (['_', ('_', 4), '_'], ['_', ('_', 5), '_']),
            compare.compare(expected, actual))

    def test_tuple_of_tuples(self):
        expected = ((1, 2), (3, 4), (5, 6))
        actual = ((1, 2), (3, 5), (5, 6))
        self.assertEqual(
            (('_', ('_', 4), '_'), ('_', ('_', 5), '_')),
            compare.compare(expected, actual))

    def test_tuple_of_lists(self):
        expected = ([1, 2], [3, 4], [5, 6])
        actual = ([1, 2], [3, 5], [5, 6])
        self.assertEqual(
            (('_', ['_', 4], '_'), ('_', ['_', 5], '_')),
            compare.compare(expected, actual))

    def test_sets_of_distinct_tuples(self):
        expected = set([(1, 2)])
        actual = set([(3, 4)])
        self.assertEqual(
            (set([(1, 2)]), set([(3, 4)])),
            compare.compare(expected, actual))

    def test_sets_of_intersecting_tuples(self):
        expected = set([(1, 2)])
        actual = set([(2, 3)])
        self.assertEqual(
            (set([(1, 2)]), set([(2, 3)])),
            compare.compare(
                expected, 
                actual,
                type_compare={'ordered' : False}))

    def test_compare_none_with_not_present(self):
        expected = {}
        actual = {'a' : None}
        self.assertEqual(
                ({'a' : compare.NotPresent}, {'a' : None}),
                compare.compare(
                    expected,
                    actual,
                    type_compare='full'))

    def test_compare_empty_list_with_empty_hash(self):
        expected = {}
        actual = []
        self.assertEqual(
            ({}, []),
            compare.compare(
                expected,
                actual))

    def test_compare_empty_hash_with_empty_list(self):
        expected = []
        actual = {}
        self.assertEqual(
            ([], {}),
            compare.compare(
                expected,
                actual))

    def test_compare_nonempty_list_with_nonempty_hash(self):
        expected = {'a': 1}
        actual = [1]
        self.assertEqual(
            ({'a': 1}, [1]),
            compare.compare(
                expected,
                actual))





import unittest

from kobold import hash_functions

class TestMakeHashable(unittest.TestCase):
    def test_str(self):
        str_to_test = 'a'
        result_str = hash_functions.make_hashable(str_to_test)
        self.assertEqual(
            str_to_test,
            result_str)
        dictionary = {result_str: result_str}
        self.assertEqual(
            result_str,
            dictionary[result_str])

    def test_list(self):
        list_to_test = ['a', 'b']
        result_list = hash_functions.make_hashable(list_to_test)
        self.assertEqual(
            list_to_test,
            result_list)
        dictionary = {result_list: result_list}
        self.assertEqual(
            result_list,
            dictionary[result_list])

    def test_dict(self):
        dict_to_test = {'a': 1}
        result_dict = hash_functions.make_hashable(dict_to_test)
        self.assertEqual(
            dict_to_test,
            result_dict)
        dictionary = {result_dict: result_dict}
        self.assertEqual(
            result_dict,
            dictionary[result_dict])

    def test_nested_dict(self):
        dict_to_test = {
            'a': {
                'a_1': {
                    'a_2': 1
                },
            }
        }
        result_dict = hash_functions.make_hashable(dict_to_test)
        self.assertEqual(
            dict_to_test,
            result_dict)
        dictionary = {result_dict: result_dict}
        self.assertEqual(
            result_dict,
            dictionary[result_dict])






import asyncio
import unittest

from kobold import (
        assertions,
        compare,
        doubles,
        swap)

class Host(object):
    def subject(self, arg, kwarg=None):
        return "original subject"

    async def subject_cr(self, arg, kwarg=None):
        return "original subject"


class TestInstallProxy(unittest.TestCase):
    def setUp(self):
        self.safe_swap = swap.SafeSwap()

    def tearDown(self):
        self.safe_swap.rollback()

    def test_proxy(self):
        self.safe_swap.install_proxy(Host, 'subject')
        host = Host()
        returned = host.subject('some_arg', kwarg='some_kwarg')
        self.assertEqual('original subject', returned)
        self.assertEqual([((host, 'some_arg'), dict(kwarg='some_kwarg'))],
                         host.subject.calls)

    def test_coroutine_proxy(self):
        host = Host()
        proxy = self.safe_swap.install_proxy(
            Host,
            'subject_cr')

        loop = asyncio.get_event_loop()
        result = loop.run_until_complete(
                host.subject_cr('1'))
        self.assertEqual(
                'original subject',
                result)
        assertions.assert_match(
            [((compare.DontCare(), '1'), {})],
            proxy.calls)



class TestSwap(unittest.TestCase):
    def setUp(self):
        self.safe_swap = swap.SafeSwap()

    def tearDown(self):
        self.safe_swap.rollback()

    def test_swap_and_rollback(self):
        stub_function = doubles.StubFunction(returns=1)
        self.safe_swap.swap(
                Host,
                'subject',
                stub_function)
        host = Host()
        self.assertEqual(1, host.subject(1, kwarg=2))
        
        self.safe_swap.rollback()

        self.assertEqual('original subject', host.subject(arg=1, kwarg=1))


    def test_default_original(self):
        routable_stub = doubles.RoutableStubFunction()
        routable_stub.add_route(
            {'kwarg': 1},
            stub_type='value',
            stub_value='new subject')
        self.safe_swap.swap(
            Host,
            'subject',
            routable_stub,
            default_original=True)
        host = Host()
        self.assertEqual(
            'new subject',
            host.subject('some_arg', kwarg=1))
        self.assertEqual(
            'original subject',
            host.subject('some_arg', kwarg=2))

    def test_default_original_coroutine(self):
        loop = asyncio.get_event_loop()
        routable_stub = doubles.RoutableStubCoroutine()
        routable_stub.add_route(
            {'kwarg': 1},
            stub_type='value',
            stub_value='new subject')
        self.safe_swap.swap(
            Host,
            'subject_cr',
            routable_stub,
            default_original=True)
        host = Host()
        self.assertEqual(
            'new subject',
            loop.run_until_complete(
                host.subject_cr('some_arg', kwarg=1)))
        self.assertEqual(
            'original subject',
            loop.run_until_complete(
                host.subject_cr('some_arg', kwarg=2)))






import kobold

def assert_response_matches(expected,
                            response,
                            type_compare=None,
                            exception_context=None):
    '''
    If the responses don't match (in the kobold.compare sense),
    raise an AssertionError
    '''

    if type_compare is None:
        type_compare = {}
    result = kobold.response.response_matches(expected,
                              response,
                              type_compare=type_compare)
    raise_if_not_match(result, exception_context=exception_context)

def assert_equal(
        expected, 
        actual, 
        type_compare=None,
        exception_context=None):
    '''
    If two data structures don't match (in the kobold.compare sense),
    raise an AssertionError
    '''

    if type_compare is None:
        type_compare = {}
    result = kobold.compare.compare(expected, 
                     actual,
                     type_compare=type_compare)
    raise_if_not_match(result, exception_context=exception_context)

assert_match = assert_equal

def raise_if_not_match(result, exception_context=None):
    if result != 'match':
        expected_diff, actual_diff = result
        assertion_text =\
            "Expected %s\nBut Got %s" %\
                    (expected_diff,
                     actual_diff)
        if exception_context is None:
            exception_text = assertion_text
        else:
            exception_text = "{}: {}".format(exception_context, assertion_text)
        raise AssertionError(exception_text)




import base64
import json
import re
import unittest

import kobold


class TestStructuredString(unittest.TestCase):
    def test_payload_match(self):
        expected_token_string = kobold.compare.StructuredString(
            regex=re.compile(r'XBL3.0 x=(.*);(.*)'),
            arguments=[
                'expected_uhs',
                'expected_payload'
            ])
        actual_token_string = 'XBL3.0 x=expected_uhs;expected_payload'
        diff = kobold.compare.compare(
            expected_token_string,
            actual_token_string)
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_payload_mismatch(self):
        expected_token_string = kobold.compare.StructuredString(
            regex=re.compile(r'XBL3.0 x=(.*);(.*)'),
            arguments=[
                'expected_uhs',
                'expected_payload'
            ])
        actual_token_string = 'XBL3.0 x=actual_uhs;actual_payload'
        diff = kobold.compare.compare(
            expected_token_string,
            actual_token_string)
        kobold.assertions.assert_match(
            (['expected_uhs', 'expected_payload'],
             ['actual_uhs', 'actual_payload']),
            diff)


class TestStructuredStringWithBase64Payload(unittest.TestCase):
    def test_full_payload_match(self):
        # When the encoded payloads are a perfect match,
        # of course kobold should identify that as a match
        uhs = 'uhs'
        expected_token_string = kobold.compare.StructuredString(
            regex=re.compile(r'XBL3.0 x=(.*);(.*)'),
            arguments=[
                uhs,
                kobold.compare.hints.Base64Hint(
                    kobold.compare.hints.JSONParsingHint(
                        {'env': 'prod',
                         'account_id': 1}
                    )
                )
            ])

        token_payload = {
            'env': 'prod',
            'account_id': 1}
        encoded_payload = base64.b64encode(
            json.dumps(token_payload).encode()).decode()

        actual_token_string = 'XBL3.0 x={};{}'.format(
            uhs,
            encoded_payload)

        diff = kobold.compare.compare(
            expected_token_string,
            actual_token_string)
        kobold.assertions.assert_match(
            'match',
            diff)


    def test_partial_payload_match(self):
        # When the encoded payloads are not a match,
        # but the payloads themselves match under the 
        # "existing" rule, kobold should identify this as a match
        # (provided the comparison itself is run as "existing")
        uhs = 'uhs'
        expected_token_string = kobold.compare.StructuredString(
            regex=re.compile(r'XBL3.0 x=(.*);(.*)'),
            arguments=[
                uhs,
                kobold.compare.hints.Base64Hint(
                    kobold.compare.hints.JSONParsingHint(
                        {'env': 'prod'}
                    )
                )
            ])

        token_payload = {
            'env': 'prod',
            'account_id': 1}
        encoded_payload = base64.b64encode(
            json.dumps(token_payload).encode()).decode()

        actual_token_string = 'XBL3.0 x={};{}'.format(
            uhs,
            encoded_payload)

        diff = kobold.compare.compare(
            expected_token_string,
            actual_token_string,
            type_compare='existing')
        kobold.assertions.assert_match(
            'match',
            diff)

    def test_wrong_payload_mismatch(self):
        # When the encoded payloads are a mismatch, 
        # and the decoded payloads mismatch on an attribute,
        # kobold should identify that as a mismatch and report
        # the attributes as the diff.

        uhs = 'uhs'
        expected_token_string = kobold.compare.StructuredString(
            regex=re.compile(r'XBL3.0 x=(.*);(.*)'),
            arguments=[
                uhs,
                kobold.compare.hints.Base64Hint(
                    kobold.compare.hints.JSONParsingHint(
                        {'env': 'prod',
                         'account_id': 2}
                    )
                )
            ])

        token_payload = {
            'env': 'prod',
            'account_id': 1}
        encoded_payload = base64.b64encode(
            json.dumps(token_payload).encode()).decode()

        actual_token_string = 'XBL3.0 x={};{}'.format(
            uhs,
            encoded_payload)

        diff = kobold.compare.compare(
            expected_token_string,
            actual_token_string)
        kobold.assertions.assert_match(
            (['_', {'account_id': 2}], ['_', {'account_id': 1}]),
            diff)





import asyncio
import fnmatch
import types
import uuid

from kobold import (
        compare,
        hash_functions)

class StubFunction(object):
    '''A Stub is a Test Double that is intended to replace a real function
       with one that just behaves as configured.  Usually, this means
       returning a predefined value.  We extend this to mean returning a
       value, raising an exception, or calling another function and returning
       the result.

       The constructor takes three optional params: returns, raises and calls.

       -If returns is supplied, the function will return the value of returns
        when called.

       -If raises is supplied, the function will raise the value of raises
        when called

       -If calls is supplied, the function will call "calls" when it is 
        called.  We will return the value that the function returns.
    '''

    def __init__(self, returns=None, raises=None, calls=None, *args, **kwargs):
        self.to_return = returns
        self.to_raise = raises
        self.to_call = calls

    def returns(self, to_return):
        '''This will make StubFunction return this value when called'''
        self.to_return = to_return
        self.to_raise = None
        self.to_call = None

    def raises(self, to_raise):
        '''This will make StubFunction raise this exception when called'''
        self.to_raise = to_raise
        self.to_return = None
        self.to_call = None

    def calls(self, to_call):
        '''This will make StubFunction call this function when called
           (and return the value that the function returns)'''
        self.to_call = to_call
        self.to_return = None
        self.to_raise = None

    def __call__(self, *args, original_reference=None, **kwargs):
        if self.to_raise:
            raise self.to_raise
        elif self.to_call:
            return self.to_call(*args, **kwargs)
        else:
            return self.to_return

    def set_original_reference(self, original_reference):
        self.original_reference = original_reference

class StubCoroutine(StubFunction):
    async def __call__(self, *args, original_reference=None, **kwargs):
        if self.to_raise:
            raise self.to_raise
        elif self.to_call:
            return await self.to_call(*args, **kwargs)
        else:
            return self.to_return


class SpyFunction(object):
    '''A Spy is a Test Double that makes a record of each time it is called,
       typically to be queried later.  SpyFunction saves the args and
       kwargs everytime it's called.  It then defers to whatever stub function
       factory is supplied (by default StubFunction).  The args and kwargs
       are passed into StubFunction, so that a Spy can specify what it's
       return behavior is'''

    def __init__(self, stub_function_factory=StubFunction, *args, **kwargs):
        self.stub_function = stub_function_factory(*args, **kwargs)
        self.calls = []

    def reset(self):
        '''Forget any calls this spy has recorded'''
        self.calls = []
        if getattr(self.stub_function, 'reset', None) is not None:
            self.stub_function.reset()

    def __call__(self, *args, original_reference=None, **kwargs):
        self.calls.append((args, kwargs))
        return self.stub_function(
            *args,
            original_reference=original_reference,
            **kwargs)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            return types.MethodType(self, instance)

    def __getattr__(self, attr):
        try:
            return super(SpyFunction, self).__getattr__(attr)
        except AttributeError:
            return getattr(self.stub_function, attr)

class SpyCoroutine(SpyFunction):
    async def __call__(self, *args, original_reference=None, **kwargs):
        self.calls.append((args, kwargs))
        if asyncio.iscoroutinefunction(self.stub_function.__call__):
            return await self.stub_function(
                *args, 
                original_reference=original_reference,
                **kwargs)
        else:
            return self.stub_function(
                *args,
                original_reference=original_reference,
                **kwargs)

def get_stub_class(
        methods_to_add, 
        coroutines_to_add=None,
        is_callable=False):
    '''Creates a class whose methods are all spies.

       methods_to_add is dictionary or a list of methods to 
       add to the class that is returned.  The methods will
       all be SpyFunctions.  If methods_to_add is a dictionary, the
       method names are the keys, and the return types are the values.
       If methods_to_add is a list, the return types are all None.

       is_callable is a flag that defaults to False.  If it is 
       set to True, we will add a __call__ method to the class
       that will make instances of the class behave like a SpyFunctions.
    '''
    if coroutines_to_add is None:
        coroutines_to_add = {}

    class StubClass(object):
        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                setattr(self, name, value)

            if is_callable:
                self.calls = []

        if is_callable:
            def __call__(self, *args, **kwargs):
                self.calls.append((args, kwargs))

    if isinstance(methods_to_add, dict):
        for (method_name, return_value) in methods_to_add.items():
            spy_function = SpyFunction(returns=return_value)
            setattr(StubClass, method_name, spy_function)
    elif isinstance(methods_to_add, list):
        for method_name in methods_to_add:
            setattr(StubClass, method_name, SpyFunction(returns=None))
    for (method_name, return_value) in coroutines_to_add.items():
        spy_coroutine = SpyCoroutine(returns=return_value)
        setattr(StubClass, method_name, spy_coroutine)

    return StubClass
                

class Route(object):
    def __init__(self, condition, stub_type, stub_value, priority=0):
        self.condition = condition
        self.stub_type = stub_type
        self.stub_value = stub_value
        self.priority = priority

    def select(self):
        return (self.condition, self.stub_type, self.stub_value)

    def __repr__(self):
        return 'Route(condition={})'.format(self.condition)




class RotatingRoute(object):
    def __init__(self, condition, routes, priority=0):
        self.condition = condition
        self.routes = routes
        self.index = 0
        self.priority = priority

    def select(self):
        stub_type, stub_value = self.routes[self.index]
        self.index = (self.index + 1) % len(self.routes)
        return (self.condition, stub_type, stub_value)

    def add_route(self, route):
        self.routes.append(route)


class RoutableStubFunction(object):

    def __init__(self, default_route=None, *args, **kwargs):
        self.routes = {}
        self.route_indexes = {}
        if (default_route is not None and 
             not isinstance(default_route, Route)):
            raise NotImplementedError(
                'default_route must be of an instance of doubles.Route')
        self.default_route = default_route
        self.host = compare.NotPresent
        self.calls_by_key = {}

    def add_route(
            self,
            condition=None,
            stub_type=None,
            stub_value=None,
            route=None,
            key=None,
            route_priority=0):
        if key is None:
            key = uuid.uuid4().hex
        if route is None:
            route = Route(
                condition,
                stub_type,
                stub_value,
                priority=route_priority)

        self.routes[key] = route

        if route.condition == 'default':
            self.default_route = route

    def reset(self):
        self.calls_by_key = {}

    def clear_routes(self):
        self.routes = {}
        self.default_route = None

    def calls_for_pattern(self, pattern):
        match_calls = []
        for key, calls in self.calls_by_key.items():
            if fnmatch.fnmatch(key, pattern):
                match_calls.extend(calls)
        return match_calls

    def set_original_reference(self, original_reference):
        self.original_reference = original_reference

    def default_original(self, original_reference):
        self.default_route = Route('default', 'callable', original_reference)

    def get_candidates(self, args, kwargs):
        candidates = []
        for key, route in self.routes.items():
            condition = route.condition
            if condition == 'default':
                continue
            
            if hash_functions.acts_like_a_hash(condition):
                if set(condition.keys()).intersection(
                        set(['args', 'kwargs', 'self'])):
                    thing_to_compare = {
                        'args' : args,
                        'kwargs' : kwargs,
                        'self': self.host}
                    condition_self = condition.get('self')
                    if (condition_self is not None and 
                            isinstance(condition_self, dict)):
                        condition['self'] =\
                            compare.hints.ObjectAttrParsingHint(
                                condition_self)
                else:
                    thing_to_compare = kwargs
            elif hash_functions.acts_like_a_list(condition):
                thing_to_compare = args
            else:
                raise Exception("Unknown condition type: %s" % type(condition))

            if compare.compare(
                    condition, 
                    thing_to_compare, 
                    type_compare='existing') == 'match':
                candidates.append((key, route))

        if len(candidates) == 0:
            if self.default_route:
                candidates.append(('default', self.default_route))

        candidates = self.filter_candidates_for_priority(candidates)

        return candidates

    def filter_candidates_for_priority(self, candidates):
        if len(candidates) > 1:
            priorities = sorted([x[1].priority for x in candidates])
            best_priority = priorities[0]
            return [x for x in candidates if x[1].priority == best_priority]
        else:
            return candidates

    def __get__(self, obj, objtype):
        self.host = obj
        return self

    def __call__(self, *args, original_reference=None, **kwargs):
        candidates = self.get_candidates(args, kwargs)

        if len(candidates) > 1:
            raise StubRoutingException("More than one route candidate for stub: %s" % candidates)

        if len(candidates) == 0:
            raise StubRoutingException(
                "No route candidates for stub {} {}.  Routes: {}".format(
                    args,
                    kwargs,
                    self.routes))

        key, route = candidates[0]
        calls_for_key = self.calls_by_key.get(key)
        if calls_for_key is None:
            calls_for_key = self.calls_by_key[key] = []

        calls_for_key.append((args, kwargs))

        (condition, stub_type, stub_value) = route.select()
        if stub_type == 'value':
            return stub_value

        if stub_type == 'callable':
            if self.host is compare.NotPresent:
                to_call = stub_value
            else:
                to_call = stub_value.__get__(
                        self.host,
                        self.host.__class__)
            return to_call(*args, **kwargs)

        if stub_type == 'exception':
            raise stub_value

        if stub_type == 'original':
            if original_reference is None:
                original_reference = self.original_reference

            if self.host is compare.NotPresent:
                to_call = original_reference
            else:
                to_call = original_reference.__get__(
                        self.host,
                        self.host.__class__)
            return to_call(*args, **kwargs)


class RoutableStubCoroutine(RoutableStubFunction):
    async def __call__(self, *args, original_reference=None, **kwargs):
        candidates = self.get_candidates(args, kwargs)

        if len(candidates) > 1:
            raise StubRoutingException("More than one route candidate for stub: %s" % candidates)

        if len(candidates) == 0:
            raise StubRoutingException(
                "No route candidates for stub: {}; {}.  \n\nRoutes: {}".format(
                    args,
                    kwargs,
                    {k: r.condition for (k, r) in self.routes.items()}))

        key, route = candidates[0]

        calls_for_key = self.calls_by_key.get(key)
        if calls_for_key is None:
            calls_for_key = self.calls_by_key[key] = []

        calls_for_key.append((args, kwargs))

        (condition, stub_type, stub_value) = route.select()
        if stub_type == 'value':
            return stub_value

        if stub_type == 'callable':
            if self.host is compare.NotPresent:
                to_call = stub_value
            else:
                to_call = stub_value.__get__(
                        self.host,
                        self.host.__class__)
            return await to_call(*args, **kwargs)

        if stub_type == 'exception':
            raise stub_value

        if stub_type == 'original':
            if original_reference is None:
                original_reference = self.original_reference
            if self.host is compare.NotPresent:
                to_call = original_reference
            else:
                to_call = original_reference.__get__(
                        self.host,
                        self.host.__class__)
            return await to_call(*args, **kwargs)


class StubRoutingException(Exception):
    '''This exception indicates that a RoutableStubFunction could
       not determine which route to follow.  A RoutableStubFunction
       should always be able to identify exactly one route to 
       follow for each function call'''
    pass


